/*
 * Copyright 2009 Andreas Borglin andreas.borglin@gmail.com
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

package borge.mobile.lib.pimp;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * Base class for PreviewHandler and RealHandler
 * Contains all common functionality
 * @author Andreas Borglin
 */
abstract class BaseHandler implements ImageConstants{
    
    protected String inputFile;
    protected ImageEffectProvider effectProvider;
    protected int width;
    protected int height;
    protected boolean initialized;
    protected boolean sendPostOp;
    protected boolean undoMode;
    protected boolean lastOpResize;
    
    // Same for all instances...
    protected static ImageSupport imageSupport;
    
    // These are simply the filetypes this framework supports.
    // There are no guarentees that the platform actually supports the associated formats
    protected static final String[] SUPPORTED_FILETYPES = {"jpg","jpeg","gif","png","bmp"};
    
    /**
     * This internal class is used to keep track of which image operations
     * the user has chosen. Once the user chooses to save the changes to file,
     * we will parse all PerformedOperations and apply them to the real image.
     */
    protected class PerformedOperation{
        
        public int operation;
        public int[] values;
        public String sValue;
        public boolean conflict = false;
        
        protected PerformedOperation(int op, int[] vals){
            operation = op;
            values = vals;
        }
        
        protected PerformedOperation(int op, int[] vals, String val){
            this(op, vals);
            sValue = val;
        }
    }
    
    protected BaseHandler(boolean undo, boolean postop) {
        undoMode = undo;
        sendPostOp = postop;
    }
    
    protected String getInputFilePath(){
        return inputFile;
    }
    
    /** Initialize the MediaProcessor and create controls */
    protected void initialize(String file, OutputStream os) throws ImageEffectException {
        
        if(effectProvider != null){
            reset();
            effectProvider.close();
            effectProvider = null;
            initialized = false;
        }
        
        // Create preview provider and point to that instance
        effectProvider = new JSR234Provider(undoMode);
        
        inputFile = file; // Save for later
        providerInit(effectProvider, inputFile, os);
        initialized = true;

    }
    
    protected void providerInit(ImageEffectProvider iep, String file, OutputStream os) throws ImageEffectException{
        
        String format = ImageInfo.getFormatFromExtension(file);
        if(format == null)
            throw new ImageEffectException("Unsupported file type");
        if(os == null)
            os = new ImageOutputStream();
        
        try{
            iep.initialize(format, file, os);
        }
        catch(SecurityException e){
            throw new ImageEffectException("Access denied for file directory");
        }
        catch(Exception e){
            e.printStackTrace();
            throw new ImageEffectException("File is not of known/supported image format");
        }
        
    }

    /**
     * Close all files and streams
     */
    protected void close(){
        effectProvider.close();
    }
    
    protected void setImageCreator(ImageCreator ic) throws ImageEffectException {
        checkState();
        effectProvider.setImageCreator(ic);
    }
    
    protected String[] getSupportedInputFormats(){
        return effectProvider.getSupportedInputFormats();
    }
    
    protected String[] getSupportedOutputFormats(){
        return effectProvider.getSupportedOutputFormats();
    }
    
    protected ImageSupport getSupportInfo(){
        return effectProvider.getImageSupport();
    }
    
    // Default - Do nothing...
    protected void postOperation(PerformedOperation po) {}
    
        /** Undo the latest step */
    protected boolean undo(){
        return effectProvider.undo();
    }
    
    /** Has there been any changes? */
    protected boolean canUndo(){
        return effectProvider.canUndo();
    }

    // Do we support this???
    protected boolean redo(){
        return effectProvider.redo();
    }
    
    /** Apply all changes to image data */
    protected void applyChanges() throws ImageEffectException {

        // For resize in preview, nothing to apply
        if(!lastOpResize)
            effectProvider.applyChanges();
        else
            lastOpResize = false;
    }
    
    protected void reset(){
        
        // Undo might be overloaded
        // Undo while possible, which will return us
        // to original state
        while(undo());
        
    }
    
    // Convenience function
    protected InputStream getCurrentImageStream(){
        return new ByteArrayInputStream(getCurrentImageData());
    }
    
    protected byte[] getCurrentImageData(){
        return effectProvider.getCurrentImageData();
    }
    
    protected void setOutputFormat(String format) throws ImageEffectException{
        effectProvider.setOutputFormat(format);
    }
    
    protected void checkState() throws ImageEffectException{
        
        if(!initialized)
            throw new ImageEffectException("Not initialized");
        
    }
    
    // Start of image operation methods
    // 1. Check that we are in correct state
    // 2. Check parameters
    // 3. Perform operation via ImageEffectProvider
    // 4. If set, call overriden postOperation with operation data
    
    // Can't handle resize generic
    // Resize are handled separately by PreviewHandler and RealHandler
    /*
    protected void resize(int newWidth, int newHeight) throws ImageEffectException;
    protected void resize(int percent) throws ImageEffectException;
    */
    
    protected void rotate(int rotation) throws ImageEffectException{
        
        checkState();
        if(rotation != ROTATE_90 && rotation != ROTATE_180 && rotation != ROTATE_270)
            throw new ImageEffectException("Value must be one of the predefined rotation mode!");

        if(!handleRotate(rotation))
            effectProvider.performOperation(ROTATE, rotation, NO_VALUE,NO_VALUE,NO_VALUE);
        
        if(sendPostOp)
            postOperation(new PerformedOperation(ROTATE,new int[]{rotation}));
    }
    // om rotate 90, 270
    // vnd p width and height..rkna ut ny size fr bilden
    // ny metod..rotateAndResize..gr i ett steg

    // For preview, we need to handle 90/270 rotation specially
    protected boolean handleRotate(int rotation)  throws ImageEffectException{
        return false;
    }
    
    protected void flip(int direction) throws ImageEffectException {
        
        checkState();
        if(direction != FLIP_HORIZONTALLY && direction != FLIP_VERTICALLY &&
                direction != ImageEffectProvider.FLIP_BOTH)
            throw new ImageEffectException("Flip direction must be one of predefined values");
        
        effectProvider.performOperation(FLIP,direction,NO_VALUE,NO_VALUE,NO_VALUE);
        
        if(sendPostOp)
            postOperation(new PerformedOperation(FLIP,new int[]{direction}));
    }
    
    protected void crop(int startX, int startY, int width, int height) throws ImageEffectException {
        
        checkState();
        effectProvider.performOperation(CROP, startX, startY, width, height);
        
        if(sendPostOp){
            // Convert to values for real image
            int x = calcRealValue(startX);
            int y = calcRealValue(startY);
            int w = calcRealValue(width);
            int h = calcRealValue(height);
            postOperation(new PerformedOperation(CROP, new int[]{x,y,w,h}));
        }
    }
    
    protected void frame(int frameWidth, int frameHeight) throws ImageEffectException {
        
        checkState();
        effectProvider.performOperation(FRAME, frameWidth, frameHeight, NO_VALUE,NO_VALUE);
        
        if(sendPostOp){
            // Convert to values for real image
            int w = calcRealValue(frameWidth);
            int h = calcRealValue(frameHeight);
            postOperation(new PerformedOperation(FRAME,new int[]{w,h}));    
        }
    }
    
    protected void setBrightness(int brightnessValue) throws ImageEffectException {
        
        checkState();
        effectProvider.performOperation(BRIGHTNESS,brightnessValue,NO_VALUE,NO_VALUE,NO_VALUE);
        
        if(sendPostOp)
            postOperation(new PerformedOperation(BRIGHTNESS,new int[]{brightnessValue}));
    }
    
    protected void setContrast(int contrastValue) throws ImageEffectException {
        
        checkState();
        effectProvider.performOperation(CONTRAST,contrastValue,NO_VALUE,NO_VALUE,NO_VALUE);
        
        if(sendPostOp)
            postOperation(new PerformedOperation(CONTRAST, new int[]{contrastValue}));
    }
    
    protected void setGamma(int gammaValue) throws ImageEffectException {
        
        checkState();
        effectProvider.performOperation(GAMMA,gammaValue,NO_VALUE,NO_VALUE,NO_VALUE);
        
        if(sendPostOp)
            postOperation(new PerformedOperation(GAMMA, new int[]{gammaValue}));
    }
    
    protected void setWhiteBalance(int index) throws ImageEffectException {
        
        checkState();
        effectProvider.performOperation(WHITEBALANCE,index,NO_VALUE,NO_VALUE,NO_VALUE);
        
        if(sendPostOp)
            postOperation(new PerformedOperation(WHITEBALANCE,new int[]{index}));
    }
    
    protected void applyEffect(int index) throws ImageEffectException {
        
        checkState();
        effectProvider.performOperation(EFFECT,index,NO_VALUE,NO_VALUE,NO_VALUE);
        
        if(sendPostOp)
            postOperation(new PerformedOperation(EFFECT, new int[]{index}));
    }
    
    private void validateIntValue(int value, int min, int max) throws ImageEffectException {
        
        if(value < min || value > max)
            throw new ImageEffectException("Value out of range: " + value);
        
    }
    
    protected boolean isValidFileType(String file){
        
        String elem = file.toLowerCase();
        
        for(int i = 0; i < SUPPORTED_FILETYPES.length; ++i){
            
            if(elem.equals(SUPPORTED_FILETYPES[i]))
                return true;
        }
        
        return false;
    }

    protected int getImageWidth() {
        return effectProvider.getImageWidth();
    }

    protected int getImageHeight() {
        return effectProvider.getImageHeight();
    }
    
    // Overriden by PreviewHandler...
    protected int calcRealValue(int value){
        return value;
    }
}
